from dataclasses import dataclass

import genanki

from memium.destination.ankiconnect.anki_model import AnkiQAModel
from memium.utils.hash_cleaned_str import hash_str_to_int
from memium.utils.markdown import md_to_html


@dataclass(frozen=True)
class AnkiQAFormatter:
    css: str

    CARD_NAME = "Ankdown QA with raw text"

    def format(self, model: AnkiQAModel) -> genanki.Note:
        return genanki.Note(
            guid=str(model.raw_prompt.scheduling_uid),
            model=self._genanki_model(model),
            fields=[
                *[md_to_html(getattr(model, field)) for field in model.formatted_field_names],
                model.raw_prompt.question,
                model.raw_prompt.answer,
            ],
            tags=model.tags,
        )

    def _genanki_model(self, model: AnkiQAModel) -> genanki.Model:
        return genanki.Model(
            model_id=hash_str_to_int(self.CARD_NAME),
            name=(self.CARD_NAME),
            fields=[
                {"name": field}
                for field in [*model.formatted_field_names, "raw_question", "raw_answer"]
            ],
            templates=self._model_template(model),
            css=self.css,
            model_type=0,
        )

    def _model_template(self, model: AnkiQAModel) -> list[dict[str, str]]:
        # TTS_QUESTION_STR = r"{{ tts en_US voices=Apple_Samantha speed=1.05:Question }}"  # noqa: ERA001
        # TTS_ANSWER_STR = r"{{ tts en_US voices=Apple_Samantha speed=1.05:Answer }}"  # noqa: ERA001

        QUESTION_PLACEHOLDER = r"{{ Question }}"
        ANSWER_PLACEHOLDER = r"{{ Answer }}"
        EXTRA_PLACEHOLDER = r"{{ Extra }}"

        placeholders = [QUESTION_PLACEHOLDER, ANSWER_PLACEHOLDER, EXTRA_PLACEHOLDER]

        for placeholder in placeholders:
            if (
                placeholder.replace(r"{{ ", "").replace(r" }}", "")
                not in model.formatted_field_names
            ):
                raise ValueError(f"Field {placeholder} is not defined in the model")

        return [
            {
                "name": self.CARD_NAME,
                "qfmt": f"""
<div class="front">{EXTRA_PLACEHOLDER}
    {QUESTION_PLACEHOLDER}
</div>""",
                "afmt": f"""
<div class="back">{EXTRA_PLACEHOLDER}
    <div class="question">
        {QUESTION_PLACEHOLDER}
    </div>
    <div class="answer">
        {ANSWER_PLACEHOLDER}
    </div>
</div>
            """,
            }
        ]



import pytest

from .hash_cleaned_str import clean_str, hash_str_to_int


def hash_cleaned_str(input_str: str) -> int:
    return hash_str_to_int(clean_str(input_str))


def test_hash_cleaned_str_should_ignore_punctuation():
    strings_should_hash_to_identical = ["this", "This"]

    punctuation = r"!().,:;/"
    strings_should_hash_to_identical += [
        f"{s}{punctuation}" for s in strings_should_hash_to_identical
    ]

    assert len({hash_str_to_int(clean_str(s)) for s in strings_should_hash_to_identical}) == 1


@pytest.mark.parametrize(("input_str", "hash_identical_str"), [("<p>Test</p>", "Test"), ("å", "å")])
def test_hash_cleaned_str_should_remove_html_tags(input_str: str, hash_identical_str: str):
    assert hash_cleaned_str(input_str) == hash_cleaned_str(hash_identical_str)


@pytest.mark.parametrize(
    ("input_str", "expected"),
    [
        ("Is <2, but >4", "is_2_but_4"),
        ("Q. ![](src). Question?", "q_src_question"),
        ("pre <img src='testsrc' /> post", "pre_testsrc_post"),
        ("\t", ""),
        ("\n", ""),
        ("""1. One\n\t2. Two""", "one_two"),
        ("""* One\n\t* Two""", "one_two"),
        ("""- One\n\t- Two""", "one_two"),
        ("[link1](blah) and [link2](blah)", "link1_and_link2"),
        ('<pre><code class="language-r">ls()</code></pre>', "ls"),
        (
            """```r
ls()
```""",
            "ls",
        ),
    ],
)
def test_str_cleaner(input_str: str, expected: str):
    assert clean_str(input_str) == expected


@pytest.mark.parametrize(
    ("markdown_str", "html"),
    [
        ('```df.select(["A", "B"])```', '<p><code>df.select(["A", "B"])</code></p>'),
        (
            """``` js
for (let i = 0; i < 4; i += 1) {
	console.log(i);
};
```""",
            """<pre><code class="language-js">for (let i = 0; i &lt; 4; i += 1) {
    console.log(i);
};
</code></pre>""",
        ),
        (
            """``` python
```""",
            """<pre><code class="language-python"></code></pre>""",
        ),
        (
            """```python
df.filter(pl.col("ID").is_in(ids_set))
```""",
            """<pre><code class="language-python">df.filter(pl.col("ID").is_in(ids_set))
</code></pre>""",
        ),
    ],
)
def test_str_cleaner_preserves_identity(markdown_str: str, html: str):
    cleaned_markdown = clean_str(markdown_str)
    cleaned_html = clean_str(html)
    assert cleaned_markdown == cleaned_html
    assert hash_cleaned_str(markdown_str) == hash_cleaned_str(html)



from collections.abc import Mapping

import pytest

from memium.source.document import Document

from ..source.prompt import QAWithDoc
from .ankiconnect.ankiconnect_gateway import (
    AnkiField,
    ImportPackage,
    SpieAnkiconnectGateway,
    UpdateModel,
)
from .ankiconnect.test_ankiconnect_gateway import MockNoteInfo
from .destination import PushPrompts
from .destination_ankiconnect import AnkiConnectDestination


@pytest.mark.parametrize(
    ("fields"),
    [
        (
            {
                "Question": AnkiField(value="MockQuestion", order=0),
                "Answer": AnkiField(value="MockAnswer", order=1),
                "raw_question": AnkiField(value="MockQuestion", order=2),
                "raw_answer": AnkiField(value="MockAnswer", order=3),
                "Extra": AnkiField(value="MockExtra", order=4),
            }
        ),
        pytest.param(
            ({"UnknownField": AnkiField(value="MockText", order=0)}), marks=pytest.mark.xfail
        ),
    ],
)
def test_ankiconnect_get_all_prompts(fields: Mapping[str, AnkiField]):
    dest = AnkiConnectDestination.dummy(
        gateway=SpieAnkiconnectGateway(note_infos=[MockNoteInfo(fields=fields)])
    )
    prompts = dest.get_all_prompts()

    assert len(prompts) == 1


def test_ankiconnect_push_prompts():
    gateway = SpieAnkiconnectGateway()
    dest = AnkiConnectDestination.dummy(gateway=gateway)
    dest.update(
        [
            PushPrompts(
                prompts=[
                    QAWithDoc.dummy(
                        question="FakeQuestion",
                        answer="FakeAnswer",
                        parent_doc=Document.dummy(tags=["anki/deck/FakeSubdeck/FakeSubSubdeck"]),
                    )
                ]
            )
        ]
    )

    expected_commands = [(ImportPackage, 1), (UpdateModel, 1)]
    import_package_command = next(
        c for c in gateway.executed_commands if isinstance(c, ImportPackage)
    )
    assert (
        import_package_command.package.decks[0].name  # type: ignore
        == "FakeBaseDeck::FakeSubdeck::FakeSubSubdeck"
    )
    assert len(import_package_command.package.decks) == 1  # type: ignore

    for command in expected_commands:
        assert (
            len([c for c in gateway.executed_commands if isinstance(c, command[0])]) == command[1]
        )



import pytest
from inline_snapshot import snapshot

from memium.source.prompt import QAPrompt, obsidian_url


def test_file_title_to_uri():
    title = "Heap Properties"
    assert obsidian_url(title, 5) == snapshot(
        "obsidian://advanced-uri?filename=Heap%20Properties&line=5"
    )


def test_should_error_on_styling():
    with pytest.raises(ValueError) as excinfo:  # noqa: PT011
        QAPrompt("Testing <p style='opacity: 1'>", "")

    assert str(excinfo.value) == snapshot("'style=' found in question")



from collections.abc import Sequence
from dataclasses import dataclass

from iterpy import Arr

from memium.source.prompt import QAPrompt, QAWithDoc


@dataclass(frozen=True)
class TitleAsAnswerProcessor:
    """Extracts prompts from a document, but does not add the document to the front of the prompt. E.g.:

    # Lion
    Q. Definition?
    A. A large African cat with a mane.

    Results in:
    Q. Term for 'A large African cat with a mane'?
    A. Lion
    """

    question_matcher: str
    reversed_question: str

    def __post_init__(self):
        if "%s" not in self.reversed_question:
            raise ValueError(
                "reversed_question must contain '%s' to be replaced with the document title"
            )

    def _expand_prompt(self, prompt: QAWithDoc) -> QAWithDoc:
        answer = prompt.prompt.answer
        if answer[-1] in ".,!?;:":
            answer = answer[:-1]

        return QAWithDoc(
            prompt=QAPrompt(
                question=self.reversed_question % answer, answer=prompt.parent_doc.source_path.stem
            ),
            parent_doc=prompt.parent_doc,
            line_nr=prompt.line_nr,
            render_parent_doc=False,
        )

    def __call__(self, prompts: Sequence[QAWithDoc]) -> Sequence[QAWithDoc]:
        definition_prompts = (
            Arr(prompts)
            .filter(lambda p: self.question_matcher in p.prompt.question)
            .map(lambda p: self._expand_prompt(p))
        )

        return [*prompts, *definition_prompts]



import logging
import re
from collections.abc import Sequence

from ..document import Document
from ..prompt import QAPrompt, QAWithDoc
from .extractor import BasePromptExtractor

log = logging.getLogger(__name__)


class QAPromptExtractor(BasePromptExtractor):
    def __init__(self, question_prefix: str, answer_prefix: str) -> None:
        self.question_prefix = question_prefix
        self.answer_prefix = answer_prefix

    def _get_first_question(self, content: str) -> str:
        question = re.findall(
            self.question_prefix + r"{0,1}\.(?:(?!A\.).)*", content, flags=re.DOTALL
        )[0]

        return question[len(self.question_prefix) + 1 :].rstrip()

    def _get_first_answer(self, content: str) -> str:
        # Have to use positive lookahead to match code-blocks
        # To ensure the last answer is matched as well, we add 2 newlines to string.
        string_padded = f"{content.rstrip()}\n\n"

        answer = re.findall(r"\n" + self.answer_prefix + r"[ \n]+\n*.+", string_padded, re.DOTALL)[
            0
        ]

        return answer[len(self.answer_prefix) + 2 :].rstrip()

    @staticmethod
    def _string_to_blocks_by_newlines(string: str) -> list[str]:
        """Break string into a list by 2+ newlines in a row."""
        return re.split(r"(\n\n)+", string)

    def _has_qa(self, string: str) -> bool:
        """Check whether a string contains a qa prompt"""
        return (
            len(
                re.findall(
                    r"^(?![:>]).*" + self.question_prefix + r"{0,1}\. ", string, flags=re.DOTALL
                )
            )
            != 0
        )

    def extract_prompts(self, document: Document) -> Sequence[QAWithDoc]:
        prompts: list[QAWithDoc] = []
        blocks = self._string_to_blocks_by_newlines(document.content)
        block_starting_line_nr = 1

        missing_answers: list[str] = []
        for block_string in blocks:
            if self._has_qa(block_string):
                question = self._get_first_question(block_string)
                try:
                    answer = self._get_first_answer(block_string)
                except IndexError:
                    missing_answers.append(f"{document.title}: {question}")
                    continue

                prompts.append(
                    QAWithDoc(
                        prompt=QAPrompt(question=question, answer=answer),
                        parent_doc=document,
                        line_nr=block_starting_line_nr,
                    )
                )

            block_lines = len(re.findall(r"\n", block_string, flags=re.DOTALL))
            block_starting_line_nr += block_lines

        if missing_answers:
            logging.warning(f"{missing_answers} is missing an answer")

        return prompts



from dataclasses import dataclass

from memium.source.prompt import QAPrompt


@dataclass(frozen=True)
class AnkiPrompt:
    prompt: QAPrompt

    source_title: str | None
    edit_url: str | None

    render_parent_doc: bool



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, TypeAlias

from memium.source.prompt import QAPrompt, QAWithDoc


@dataclass(frozen=True)
class DestinationPrompt:
    prompt: QAPrompt

    extra: str

    # Destinations can have their own ID-scheme. E.g. Anki assigns a Note id at creation time, which
    # is independent of content. This ID is used to delete or update the note.
    destination_id: str

    @staticmethod
    def dummy(
        prompt: QAWithDoc, extra: str = "DummyExtra", destination_id: str = "DummyDestinationID"
    ) -> "DestinationPrompt":
        return DestinationPrompt(prompt=prompt.prompt, extra=extra, destination_id=destination_id)


@dataclass(frozen=True)
class PushPrompts:
    prompts: Sequence[QAWithDoc]


@dataclass(frozen=True)
class DeletePrompts:
    prompts: Sequence[DestinationPrompt]


PromptDestinationCommand: TypeAlias = PushPrompts | DeletePrompts


class PromptDestination(Protocol):
    def get_all_prompts(self) -> Sequence[DestinationPrompt]: ...

    def update(self, commands: Sequence[PromptDestinationCommand]) -> None: ...



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Generic, TypeVar

from memium.source.prompt import QAWithDoc

from .destination.destination import DestinationPrompt

K = TypeVar("K")
T = TypeVar("T")
S = TypeVar("S")


@dataclass(frozen=True)
class GeneralSyncer(Generic[K, T, S]):
    source: Mapping[K, T]
    destination: Mapping[K, S]

    def only_in_source(self) -> Sequence[T]:
        return [value for key, value in self.source.items() if key not in self.destination]

    def only_in_destination(self) -> Sequence[S]:
        return [value for key, value in self.destination.items() if key not in self.source]


class PromptDiffDeterminer:
    def to_delete(
        self, source_prompts: Sequence[QAWithDoc], destination_prompts: Sequence[DestinationPrompt]
    ) -> Sequence[DestinationPrompt]:
        # Only delete prompts whose content have changed. This essentially resets their scheduling.
        prompts_to_delete = GeneralSyncer(
            source={prompt.scheduling_str: prompt for prompt in source_prompts},
            destination={
                dest_prompt.prompt.scheduling_uid_str: dest_prompt
                for dest_prompt in destination_prompts
            },
        ).only_in_destination()

        return prompts_to_delete



from collections.abc import Sequence

import pytest

from memium.source.document import Document
from memium.source.prompt import QAWithDoc

from .destination.destination import DestinationPrompt
from .diff_determiner import GeneralSyncer, PromptDiffDeterminer


@pytest.fixture
def diff_determiner() -> PromptDiffDeterminer:
    return PromptDiffDeterminer()


def test_diff_determiner():
    syncer = GeneralSyncer(source={"a": 1, "b": 2}, destination={"b": "2", "c": "3"})

    assert syncer.only_in_source() == [1]
    assert syncer.only_in_destination() == ["3"]


from dataclasses import dataclass


@dataclass(frozen=True)
class DiffDeterminerExample:
    example_title: str
    source_prompts: Sequence[QAWithDoc]
    destination_prompts: Sequence[DestinationPrompt]
    should_delete_prompts: Sequence[DestinationPrompt]


@pytest.mark.parametrize(
    "example",
    [
        DiffDeterminerExample(
            source_prompts=[
                QAWithDoc.dummy(question="a", answer="a"),
                QAWithDoc.dummy(question="b", answer="b"),
            ],
            destination_prompts=[
                DestinationPrompt.dummy(
                    QAWithDoc.dummy(question="b", answer="b"), destination_id="2"
                ),
                DestinationPrompt.dummy(
                    QAWithDoc.dummy(question="c", answer="c"), destination_id="3"
                ),
            ],
            should_delete_prompts=[
                DestinationPrompt.dummy(
                    QAWithDoc.dummy(question="c", answer="c"), destination_id="3"
                )
            ],
            example_title="Should delete 'c' only in destination",
        ),
        (
            DiffDeterminerExample(
                source_prompts=[QAWithDoc.dummy(parent_doc=Document.dummy(tags=["#NewTag"]))],
                destination_prompts=[
                    DestinationPrompt.dummy(
                        QAWithDoc.dummy(parent_doc=Document.dummy(tags=["OldTag"])),
                        destination_id="1",
                    )
                ],
                should_delete_prompts=[],
                example_title="Updated tags should not delete",
            )
        ),
    ],
    ids=lambda example: example.example_title,
)
def test_prompt_diff_determiner(
    diff_determiner: PromptDiffDeterminer, example: DiffDeterminerExample
):
    to_delete = diff_determiner.to_delete(
        source_prompts=example.source_prompts, destination_prompts=example.destination_prompts
    )
    assert to_delete == example.should_delete_prompts



import logging
import sys
import time
from datetime import datetime
from functools import partial
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import Annotated, Optional

import typer
from iterpy import Arr

from memium.destination.ankiconnect.anki_converter import AnkiPromptConverter
from memium.destination.ankiconnect.anki_formatter import AnkiQAFormatter
from memium.destination.ankiconnect.ankiconnect_gateway import ANKICONNECT_URL, AnkiConnectGateway
from memium.destination.destination import DeletePrompts, PushPrompts
from memium.destination.destination_ankiconnect import AnkiConnectDestination
from memium.destination.destination_dryrun import DryRunDestination
from memium.diff_determiner import PromptDiffDeterminer
from memium.environment import host_input_dir, in_docker
from memium.raw_processors.title_as_answer import TitleAsAnswerProcessor
from memium.source.document_source import MarkdownDocumentSource
from memium.source.extractors.extractor_qa import QAPromptExtractor
from memium.source.extractors.extractor_table import TableExtractor
from memium.source.prompt_source import DocumentPromptSource

log = logging.getLogger(__name__)

app = typer.Typer()


def main(root_deck: str, input_dir: Path, max_deletions_per_run: int, dry_run: bool):
    # Setup gateway as first step. If Anki is not running, no need to parse all the prompts.
    gateway = AnkiConnectGateway(
        ankiconnect_url=ANKICONNECT_URL,
        root_deck=root_deck,
        tmp_read_dir=host_input_dir() if in_docker() else input_dir,
        tmp_write_dir=input_dir,
        max_deletions_per_run=max_deletions_per_run,
        max_wait_seconds=3600,
    )

    # Get the inputs
    qa_extractor = QAPromptExtractor(question_prefix="Q.", answer_prefix="A.")
    source_prompts = DocumentPromptSource(
        document_ingester=MarkdownDocumentSource(directory=input_dir),
        prompt_extractors=[qa_extractor, TableExtractor()],
    ).get_prompts()

    # Apply transformations
    transformed_source_prompts = (
        Arr([source_prompts])
        .map(
            TitleAsAnswerProcessor(
                question_matcher="Definition?", reversed_question="Term for '%s'?"
            )
        )
        .map(
            TitleAsAnswerProcessor(
                question_matcher="Use when?", reversed_question="What should we use for '%s'?"
            )
        )
        .map(
            TitleAsAnswerProcessor(
                question_matcher="Avoid when?", reversed_question="What should we avoid when '%s'?"
            )
        )
        .map(
            TitleAsAnswerProcessor(
                question_matcher="Antonym?", reversed_question="What is the antonym of '%s'?"
            )
        )
        .flatten()
        .to_list()
    )

    # Determine diff
    dest_class = AnkiConnectDestination if not dry_run else DryRunDestination
    destination = dest_class(
        gateway=gateway,
        prompt_converter=AnkiPromptConverter(root_deck=root_deck),
        formatter=AnkiQAFormatter(
            Path("memium/destination/ankiconnect/default_styling.css").read_text()
        ),
    )
    delete_prompts = PromptDiffDeterminer().to_delete(
        source_prompts=transformed_source_prompts, destination_prompts=destination.get_all_prompts()
    )

    # Synchronise
    destination.update(
        commands=[DeletePrompts(delete_prompts), PushPrompts(transformed_source_prompts)]
    )


@app.command()
def cli(
    input_dir: Annotated[
        Path,
        typer.Option(
            help="Where to extract prompts from.",
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
            writable=False,
        ),
    ],
    watch_seconds: Annotated[
        Optional[int],  # noqa: UP007
        typer.Option(help="Keep running, updating Anki deck every [ARG] seconds"),
    ] = None,
    deck_name: Annotated[
        str, typer.Option(help="Anki path to deck, e.g. 'Parent deck::Child deck'")
    ] = "Memium",
    max_deletions_per_run: Annotated[
        int,
        typer.Option(
            help="Maximum number of cards to delete per sync to avoid unintentional deletions. If exceeded, raises error."
        ),
    ] = 50,
    dry_run: Annotated[
        bool, typer.Option(help="Don't update via AnkiConnect, just log what would happen")
    ] = False,
    skip_sync: Annotated[
        bool, typer.Option(help="Skip all syncing, useful for smoketesting of the interface")
    ] = False,
):
    start_time = datetime.now()
    config_dir = input_dir / ".memium"
    config_dir.mkdir(exist_ok=True)

    date = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    log_path = config_dir / f"{date}-memium.log"

    logging.basicConfig(
        handlers=[logging.StreamHandler(sys.stdout), logging.FileHandler(log_path)],
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
    )

    if skip_sync:
        log.info("Skipping sync")
        return

    try:
        pkg_version = version("memium")
        log.info(f"Running memium version {pkg_version}")
    except PackageNotFoundError:
        log.info("Running memium from local source")
    log.info(f"Logging to {log_path}")

    # The watching logic requires having a "core" which can terminate.
    # Alternatively, we could do a recursive call, but that would result in
    # an infinitely growing stack.
    main_fn = partial(
        main,
        root_deck=deck_name,
        input_dir=input_dir,
        max_deletions_per_run=max_deletions_per_run,
        dry_run=dry_run,
    )
    main_fn()

    log.info(f"Logged to {log_path}")

    if watch_seconds:
        log.info(
            f"Sync complete in {(datetime.now() - start_time).total_seconds()} seconds, sleeping for {watch_seconds} seconds"
        )
        time.sleep(watch_seconds)
        main_fn()


if __name__ == "__main__":
    app()



import re
from collections.abc import Sequence
from typing import Protocol

from markdown import Markdown


class MarkdownParser(Protocol):
    def __call__(self, markdown: str) -> str: ...


def md_to_html(markdown: str) -> str:
    return Markdown(
        output_format="html", extensions=["legacy_em", "fenced_code", "tables", "nl2br"]
    ).convert(markdown)


def get_terms_surrounded_by_underscores(string: str) -> Sequence[str]:
    wikilinks = re.findall(r"_[\w\s\(\)]+?_", string)
    return [link.replace("_", "") for link in wikilinks]
